from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from strawberry.exceptions.exception import StrawberryException
from strawberry.exceptions.utils.source_finder import SourceFinder

if TYPE_CHECKING:
    from strawberry.exceptions.exception_source import ExceptionSource

    from strawberry_django.fields.filter_order import FilterOrderFieldResolver


class MissingFieldArgumentError(StrawberryException):
    def __init__(self, field_name: str, resolver: FilterOrderFieldResolver):
        self.function = resolver.wrapped_func

        self.message = f'Missing required argument "{field_name}" in "{resolver.name}"'
        self.rich_message = (
            f'[bold red]Missing argument [underline]"{field_name}" for field '
            f"`[underline]{resolver.name}[/]`"
        )
        self.annotation_message = "field missing argument"

        super().__init__(self.message)

    @cached_property
    def exception_source(self) -> ExceptionSource | None:  # pragma: no cover
        source_finder = SourceFinder()

        return source_finder.find_function_from_object(self.function)  # type: ignore


class ForbiddenFieldArgumentError(StrawberryException):
    def __init__(self, resolver: FilterOrderFieldResolver, arguments: list[str]):
        self.extra_arguments = arguments
        self.function = resolver.wrapped_func
        self.argument_name = arguments[0]

        self.message = (
            f'Found disallowed {self.extra_arguments_str} in field "{resolver.name}"'
        )
        self.rich_message = (
            f"Found disallowed {self.extra_arguments_str} in "
            f"`[underline]{resolver.name}[/]`"
        )
        self.suggestion = "To fix this error, remove offending argument(s)"

        self.annotation_message = "forbidden field argument"

        super().__init__(self.message)

    @property
    def extra_arguments_str(self) -> str:
        arguments = self.extra_arguments

        if len(arguments) == 1:
            return f'argument "{arguments[0]}"'

        head = ", ".join(arguments[:-1])
        return f'arguments "{head}" and "{arguments[-1]}"'

    @cached_property
    def exception_source(self) -> ExceptionSource | None:  # pragma: no cover
        source_finder = SourceFinder()

        return source_finder.find_argument_from_object(
            self.function,  # type: ignore
            self.argument_name,
        )



import abc
import contextlib
import contextvars
import copy
import dataclasses
import enum
import functools
import inspect
from collections.abc import Callable, Hashable, Iterable
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)

import strawberry
from asgiref.sync import sync_to_async
from django.core.exceptions import PermissionDenied
from django.db.models import Model, QuerySet
from graphql.pyutils import AwaitableOrValue
from strawberry import relay, schema_directive
from strawberry.extensions.field_extension import (
    AsyncExtensionResolver,
    FieldExtension,
    SyncExtensionResolver,
)
from strawberry.schema_directive import Location
from strawberry.types.base import StrawberryList, StrawberryOptional
from strawberry.types.field import StrawberryField
from strawberry.types.info import Info
from strawberry.types.union import StrawberryUnion
from typing_extensions import Literal, assert_never

from strawberry_django.auth.utils import aget_current_user, get_current_user
from strawberry_django.fields.types import OperationInfo, OperationMessage
from strawberry_django.pagination import OffsetPaginated
from strawberry_django.resolvers import django_resolver

from .utils.query import filter_for_user
from .utils.typing import UserType

if TYPE_CHECKING:
    from strawberry.django.context import StrawberryDjangoContext

    from strawberry_django.fields.field import StrawberryDjangoField


_M = TypeVar("_M", bound=Model)


@functools.lru_cache
def _get_user_or_anonymous_getter() -> Optional[Callable[[UserType], UserType]]:
    try:
        from .integrations.guardian import get_user_or_anonymous
    except (ImportError, RuntimeError):  # pragma: no cover
        return None

    return get_user_or_anonymous


@dataclasses.dataclass
class PermContext:
    is_safe_list: list[bool] = dataclasses.field(default_factory=list)
    checkers: list["HasPerm"] = dataclasses.field(default_factory=list)

    def __copy__(self):
        return self.__class__(
            is_safe_list=self.is_safe_list[:],
            checkers=self.checkers[:],
        )

    @property
    def is_safe(self):
        return bool(self.is_safe_list and all(self.is_safe_list))


perm_context: contextvars.ContextVar[PermContext] = contextvars.ContextVar(
    "perm-safe",
    default=PermContext(),  # noqa: B039
)


@contextlib.contextmanager
def with_perm_checker(checker: "HasPerm"):
    context = copy.copy(perm_context.get())
    context.checkers.append(checker)
    token = perm_context.set(context)
    try:
        yield
    finally:
        perm_context.reset(token)


def set_perm_safe(value: bool):
    perm_context.get().is_safe_list.append(value)


def filter_with_perms(qs: QuerySet[_M], info: Info) -> QuerySet[_M]:
    context = perm_context.get()
    if not context.checkers or context.is_safe:
        return qs

    if isinstance(qs, list):
        # return sliced queryset as-is
        return qs

    # Do not do anything is results are cached
    if qs._result_cache is not None:  # type: ignore
        set_perm_safe(False)
        return qs

    user = cast("StrawberryDjangoContext", info.context).request.user
    # If the user is anonymous, we can't filter object permissions for it
    if user.is_anonymous:
        set_perm_safe(False)
        return qs.none()

    for check in context.checkers:
        if check.target != PermTarget.RETVAL:
            continue

        qs = filter_for_user(
            qs,
            user,
            [p.perm for p in check.perms],
            any_perm=check.any_perm,
            with_superuser=check.with_superuser,
        )

    set_perm_safe(True)
    return qs


@overload
def get_with_perms(
    pk: strawberry.ID,
    info: Info,
    *,
    required: Literal[True],
    model: type[_M],
    key_attr: str = ...,
) -> _M: ...


@overload
def get_with_perms(
    pk: strawberry.ID,
    info: Info,
    *,
    required: bool = ...,
    model: type[_M],
    key_attr: str = ...,
) -> Optional[_M]: ...


@overload
def get_with_perms(
    pk: relay.GlobalID,
    info: Info,
    *,
    required: Literal[True],
    model: type[_M],
    key_attr: str = ...,
) -> _M: ...


@overload
def get_with_perms(
    pk: relay.GlobalID,
    info: Info,
    *,
    required: bool = ...,
    model: type[_M],
    key_attr: str = ...,
) -> Optional[_M]: ...


@overload
def get_with_perms(
    pk: relay.GlobalID,
    info: Info,
    *,
    required: Literal[True],
    key_attr: str = ...,
) -> Any: ...


@overload
def get_with_perms(
    pk: relay.GlobalID,
    info: Info,
    *,
    required: bool = ...,
    key_attr: str = ...,
) -> Optional[Any]: ...


def get_with_perms(
    pk,
    info,
    *,
    required=False,
    model=None,
    key_attr="pk",
):
    if isinstance(pk, relay.GlobalID):
        instance = pk.resolve_node_sync(info, required=required, ensure_type=model)
    else:
        assert model
        instance = model._default_manager.get(**{key_attr: pk})

    if instance is None:
        return None

    context = perm_context.get()
    if not context.checkers or context.is_safe:
        return instance

    user = cast("StrawberryDjangoContext", info.context).request.user
    if user and (get_user_or_anonymous := _get_user_or_anonymous_getter()) is not None:
        user = get_user_or_anonymous(user)

    for check in context.checkers:
        f = any if check.any_perm else all
        checker = check.obj_perm_checker(info, user)
        if not f(checker(p, instance) for p in check.perms):
            raise PermissionDenied(check.message)

    return instance


_return_condition = """\
When the condition fails, the following can be returned (following this priority):
1) `OperationInfo`/`OperationMessage` if those types are allowed at the return type
2) `null` in case the field is not mandatory (e.g. `String` or `[String]`)
3) An empty list in case the field is a list (e.g. `[String]!`)
4) An empty `Connection` in case the return type is a relay connection
2) Otherwise, an error will be raised
"""


def _desc(desc):
    return f"{desc}\n\n{_return_condition.strip()}"


class DjangoNoPermission(Exception):  # noqa: N818
    """Raise to identify that the user doesn't have perms for a given retval."""


class DjangoPermissionExtension(FieldExtension, abc.ABC):
    """Base django permission extension."""

    DEFAULT_ERROR_MESSAGE: ClassVar[str] = "User does not have permission."
    SCHEMA_DIRECTIVE_LOCATIONS: ClassVar[list[Location]] = [Location.FIELD_DEFINITION]
    SCHEMA_DIRECTIVE_DESCRIPTION: ClassVar[Optional[str]] = None

    def __init__(
        self,
        *,
        message: Optional[str] = None,
        use_directives: bool = True,
        fail_silently: bool = True,
    ):
        super().__init__()
        self.message = message if message is not None else self.DEFAULT_ERROR_MESSAGE
        self.fail_silently = fail_silently
        self.use_directives = use_directives

    def apply(self, field: StrawberryField) -> None:  # pragma: no cover
        if self.use_directives:
            directive = self.schema_directive
            # Avoid interfaces duplicating the directives
            if directive not in field.directives:
                field.directives.append(self.schema_directive)

    @functools.cached_property
    def schema_directive(self) -> object:
        key = "__strawberry_directive_type__"
        directive_class = getattr(self.__class__, key, None)

        if directive_class is None:

            @schema_directive(
                name=self.__class__.__name__,
                locations=self.SCHEMA_DIRECTIVE_LOCATIONS,
                description=self.SCHEMA_DIRECTIVE_DESCRIPTION,
                repeatable=True,
            )
            class AutoDirective: ...

            directive_class = AutoDirective

        return directive_class()

    @django_resolver(qs_hook=None)
    def resolve(
        self,
        next_: SyncExtensionResolver,
        source: Any,
        info: Info,
        **kwargs: dict[str, Any],
    ) -> Any:
        user = get_current_user(info)

        if (
            user
            and (get_user_or_anonymous := _get_user_or_anonymous_getter()) is not None
        ):
            user = get_user_or_anonymous(user)

        # make sure the user is loaded
        user.is_authenticated  # noqa: B018

        try:
            retval = self.resolve_for_user(
                functools.partial(next_, source, info, **kwargs),
                user,
                info=info,
                source=source,
            )
        except DjangoNoPermission as e:
            retval = self.handle_no_permission(e, info=info)

        return retval

    async def resolve_async(
        self,
        next_: AsyncExtensionResolver,
        source: Any,
        info: Info,
        **kwargs: dict[str, Any],
    ) -> Any:
        user = await aget_current_user(info)

        try:
            from .integrations.guardian import get_user_or_anonymous
        except (ImportError, RuntimeError):  # pragma: no cover
            pass
        else:
            user = user and await sync_to_async(get_user_or_anonymous)(user)

        # make sure the user is loaded
        await sync_to_async(getattr)(user, "is_anonymous")

        try:
            retval = self.resolve_for_user(
                functools.partial(next_, source, info, **kwargs),
                user,
                info=info,
                source=source,
            )
            while inspect.isawaitable(retval):
                retval = await retval
        except DjangoNoPermission as e:
            retval = self.handle_no_permission(e, info=info)

        return retval

    def handle_no_permission(self, exception: BaseException, *, info: Info):
        if not self.fail_silently:
            raise PermissionDenied(self.message) from exception

        ret_type = info.return_type

        if isinstance(ret_type, StrawberryOptional):
            ret_type = ret_type.of_type
            is_optional = True
        else:
            is_optional = False

        if isinstance(ret_type, StrawberryUnion):
            ret_types = []
            for type_ in ret_type.types:
                ret_types.append(ret_type)

                if not isinstance(type_, type):
                    continue

                if issubclass(type_, OperationInfo):
                    return type_(
                        messages=[
                            OperationMessage(
                                kind=OperationMessage.Kind.PERMISSION,
                                message=self.message,
                                field=info.field_name,
                            ),
                        ],
                    )

                if issubclass(type_, OperationMessage):
                    return type_(
                        kind=OperationMessage.Kind.PERMISSION,
                        message=self.message,
                        field=info.field_name,
                    )
        else:
            ret_types = [ret_type]

        if is_optional:
            return None

        if isinstance(ret_type, StrawberryList):
            return []

        if isinstance(ret_type, type) and issubclass(ret_type, OffsetPaginated):
            django_model = cast("StrawberryDjangoField", info._field).django_model
            assert django_model
            return django_model._default_manager.none()

        # If it is a Connection, try to return an empty connection, but only if
        # it is the only possibility available...
        for ret_possibility in ret_types:
            if isinstance(ret_possibility, type) and issubclass(
                ret_possibility,
                relay.Connection,
            ):
                return []

        # In last case, raise an error
        raise PermissionDenied(self.message) from exception

    @abc.abstractmethod
    def resolve_for_user(  # pragma: no cover
        self,
        resolver: Callable,
        user: Optional[UserType],
        *,
        info: Info,
        source: Any,
    ) -> AwaitableOrValue[Any]: ...


class IsAuthenticated(DjangoPermissionExtension):
    """Mark a field as only resolvable by authenticated users."""

    DEFAULT_ERROR_MESSAGE: ClassVar[str] = "User is not authenticated."
    SCHEMA_DIRECTIVE_DESCRIPTION: ClassVar[Optional[str]] = _desc(
        "Can only be resolved by authenticated users.",
    )

    @django_resolver(qs_hook=None)
    def resolve_for_user(
        self,
        resolver: Callable,
        user: Optional[UserType],
        *,
        info: Info,
        source: Any,
    ):
        if user is None or not user.is_authenticated or not user.is_active:
            raise DjangoNoPermission

        return resolver()


class IsStaff(DjangoPermissionExtension):
    """Mark a field as only resolvable by staff users."""

    DEFAULT_ERROR_MESSAGE: ClassVar[str] = "User is not a staff member."
    SCHEMA_DIRECTIVE_DESCRIPTION: ClassVar[Optional[str]] = _desc(
        "Can only be resolved by staff users.",
    )

    @django_resolver(qs_hook=None)
    def resolve_for_user(
        self,
        resolver: Callable,
        user: Optional[UserType],
        *,
        info: Info,
        source: Any,
    ):
        if (
            user is None
            or not user.is_authenticated
            or not getattr(user, "is_staff", False)
        ):
            raise DjangoNoPermission

        return resolver()


class IsSuperuser(DjangoPermissionExtension):
    """Mark a field as only resolvable by superuser users."""

    DEFAULT_ERROR_MESSAGE: ClassVar[str] = "User is not a superuser."
    SCHEMA_DIRECTIVE_DESCRIPTION: ClassVar[Optional[str]] = _desc(
        "Can only be resolved by superuser users.",
    )

    @django_resolver(qs_hook=None)
    def resolve_for_user(
        self,
        resolver: Callable,
        user: Optional[UserType],
        *,
        info: Info,
        source: Any,
    ):
        if (
            user is None
            or not user.is_authenticated
            or not getattr(user, "is_superuser", False)
        ):
            raise DjangoNoPermission

        return resolver()


@strawberry.input(description="Permission definition for schema directives.")
class PermDefinition:
    """Permission definition.

    Attributes
    ----------
        app:
            The app to which we are requiring permission.
        permission:
            The permission itself

    """

    app: Optional[str] = strawberry.field(
        description=(
            "The app to which we are requiring permission. If this is "
            "empty that means that we are checking the permission directly."
        ),
    )
    permission: Optional[str] = strawberry.field(
        description=(
            "The permission itself. If this is empty that means that we "
            "are checking for any permission for the given app."
        ),
    )

    @classmethod
    def from_perm(cls, perm: str):
        parts = perm.split(".")
        if len(parts) != 2:  # noqa: PLR2004
            raise TypeError(
                "Permissions need to be defined as `app_label.perm`, `app_label.`"
                " or `.perm`",
            )
        return cls(
            app=parts[0].strip() or None,
            permission=parts[1].strip() or None,
        )

    @property
    def perm(self):
        return f"{self.app or ''}.{self.permission or ''}".strip(".")

    def __eq__(self, other: object):
        if not isinstance(other, PermDefinition):
            return NotImplemented

        return self.perm == other.perm

    def __hash__(self):
        return hash((self.__class__, self.perm))


class PermTarget(enum.IntEnum):
    """Permission location."""

    GLOBAL = enum.auto()
    SOURCE = enum.auto()
    RETVAL = enum.auto()


def _default_perm_checker(info: Info, user: UserType):
    def perm_checker(perm: PermDefinition) -> bool:
        return (
            user.has_perm(perm.perm)  # type: ignore
            if perm.permission
            else user.has_module_perms(cast("str", perm.app))  # type: ignore
        )

    return perm_checker


def _default_obj_perm_checker(info: Info, user: UserType):
    def perm_checker(perm: PermDefinition, obj: Any) -> bool:
        # Check global perms first, then object specific
        return user.has_perm(perm.perm) or user.has_perm(  # type: ignore
            perm.perm,
            obj=obj,
        )

    return perm_checker


class HasPerm(DjangoPermissionExtension):
    """Defines permissions required to access the given object/field.

    Given a `app` name, the user can access the decorated object/field
    if he has any of the permissions defined in this directive.

    Examples
    --------
        To indicate that a mutation can only be done by someone who
        has "product.add_product" perm in the django system:

        >>> @strawberry.type
        ... class Query:
        ...     @strawberry.mutation(directives=[HasPerm("product.add_product")])
        ...     def create_product(self, name: str) -> ProductType:
        ...         ...

    Attributes
    ----------
        perms:
            Perms required to access this app.
        any_perm:
            If any perm or all perms are required to resolve the object/field.
        target:
            The target to check for permissions. Use `HasSourcePerm` or
            `HasRetvalPerm` as a shortcut for this.
        with_anonymous:
            If we should optimize the permissions check and consider an anonymous
            user as not having any permissions. This is true by default, which means
            that anonymous users will not trigger has_perm checks.
        with_superuser:
            If we should optimize the permissions check and consider a superuser
            as having permissions foe everything. This is false by default to avoid
            returning unexpected results. Setting this to true will avoid triggering
            has_perm checks.

    """

    DEFAULT_TARGET: ClassVar[PermTarget] = PermTarget.GLOBAL
    DEFAULT_ERROR_MESSAGE: ClassVar[str] = (
        "You don't have permission to access this app."
    )
    SCHEMA_DIRECTIVE_DESCRIPTION: ClassVar[Optional[str]] = _desc(
        "Will check if the user has any/all permissions to resolve this.",
    )

    def __init__(
        self,
        perms: Union[list[str], str],
        *,
        message: Optional[str] = None,
        use_directives: bool = True,
        fail_silently: bool = True,
        target: Optional[PermTarget] = None,
        any_perm: bool = True,
        perm_checker: Optional[
            Callable[[Info, UserType], Callable[[PermDefinition], bool]]
        ] = None,
        obj_perm_checker: Optional[
            Callable[[Info, UserType], Callable[[PermDefinition, Any], bool]]
        ] = None,
        with_anonymous: bool = True,
        with_superuser: bool = False,
    ):
        super().__init__(
            message=message,
            use_directives=use_directives,
            fail_silently=fail_silently,
        )

        if isinstance(perms, str):
            perms = [perms]

        if not perms:
            raise TypeError(f"At least one perm is required for {self!r}")

        self.perms: tuple[PermDefinition, ...] = tuple(
            PermDefinition.from_perm(p) if isinstance(p, str) else p for p in perms
        )

        assert all(isinstance(p, PermDefinition) for p in self.perms)
        self.target = target if target is not None else self.DEFAULT_TARGET
        self.permissions = perms
        self.any_perm = any_perm
        self.perm_checker = (
            perm_checker if perm_checker is not None else _default_perm_checker
        )
        self.obj_perm_checker = (
            obj_perm_checker
            if obj_perm_checker is not None
            else _default_obj_perm_checker
        )
        self.with_anonymous = with_anonymous
        self.with_superuser = with_superuser

    @functools.cached_property
    def schema_directive(self) -> object:
        key = "__strawberry_directive_class__"
        directive_class = getattr(self.__class__, key, None)

        if directive_class is None:

            @schema_directive(
                name=self.__class__.__name__,
                locations=self.SCHEMA_DIRECTIVE_LOCATIONS,
                description=self.SCHEMA_DIRECTIVE_DESCRIPTION,
                repeatable=True,
            )
            class AutoDirective:
                permissions: list[PermDefinition] = strawberry.field(
                    description="Required perms to access this resource.",
                    default_factory=list,
                )
                any: bool = strawberry.field(
                    description="If any or all perms listed are required.",
                    default=True,
                )

            directive_class = AutoDirective

        return directive_class(
            permissions=list(self.perms),
            any=self.any_perm,
        )

    @django_resolver(qs_hook=None)
    def resolve_for_user(
        self,
        resolver: Callable,
        user: Optional[UserType],
        *,
        info: Info,
        source: Any,
    ):
        if user is None or (self.with_anonymous and user.is_anonymous):
            raise DjangoNoPermission

        if (
            self.with_superuser
            and user.is_active
            and getattr(user, "is_superuser", False)
        ):
            return resolver()

        return self.resolve_for_user_with_perms(
            resolver,
            user,
            info=info,
            source=source,
        )

    def resolve_for_user_with_perms(
        self,
        resolver: Callable,
        user: Optional[UserType],
        *,
        info: Info,
        source: Any,
    ):
        if user is None:
            raise DjangoNoPermission

        if self.target == PermTarget.GLOBAL:
            if not self._has_perm(source, user, info=info):
                raise DjangoNoPermission

            retval = resolver()
        elif self.target == PermTarget.SOURCE:
            # Just call _resolve_obj, it will raise DjangoNoPermission
            # if the user doesn't have permission for it
            self._resolve_obj(source, user, source, info=info)
            retval = resolver()
        elif self.target == PermTarget.RETVAL:
            with with_perm_checker(self):
                obj = resolver()
                retval = self._resolve_obj(source, user, obj, info=info)
        else:
            assert_never(self.target)

        return retval

    def _get_cache(
        self,
        info: Info,
        user: UserType,
    ) -> dict[tuple[Hashable, ...], bool]:
        cache_key = "_strawberry_django_permissions_cache"

        cache = getattr(user, cache_key, None)
        if cache is None:
            cache = {}
            setattr(user, cache_key, cache)

        return cache

    def _has_perm(
        self,
        source: Any,
        user: UserType,
        *,
        info: Info,
    ) -> bool:
        cache = self._get_cache(info, user)

        # Maybe the result ended up in the cache in the meantime
        cache_key = (self.perms, self.any_perm)
        if cache_key in cache:
            return cache[cache_key]

        f = any if self.any_perm else all
        checker = self.perm_checker(info, user)
        has_perm = f(checker(p) for p in self.perms)
        cache[cache_key] = has_perm

        return has_perm

    def _resolve_obj(
        self,
        source: Any,
        user: UserType,
        obj: Any,
        *,
        info: Info,
    ) -> Any:
        context = perm_context.get()
        if context.is_safe:
            return obj

        if isinstance(obj, Iterable):
            return list(self._resolve_iterable_obj(source, user, obj, info=info))

        cache = self._get_cache(info, user)
        cache_key = (self.perms, self.any_perm, obj)
        has_perm = cache.get(cache_key)

        if has_perm is None:
            if isinstance(obj, OperationInfo):
                has_perm = True
            else:
                f = any if self.any_perm else all
                checker = self.obj_perm_checker(info, user)
                has_perm = f(checker(p, obj) for p in self.perms)

            cache[cache_key] = has_perm

        if not has_perm:
            raise DjangoNoPermission

        return obj

    def _resolve_iterable_obj(
        self,
        source: Any,
        user: UserType,
        objs: Iterable[Any],
        *,
        info: Info,
    ) -> Any:
        cache = self._get_cache(info, user)
        f = any if self.any_perm else all
        checker = self.obj_perm_checker(info, user)

        for obj in objs:
            cache_key = (self.perms, self.any_perm, obj)
            has_perm = cache.get(cache_key)

            if has_perm is None:
                if isinstance(obj, OperationInfo):
                    has_perm = True
                else:
                    has_perm = f(checker(p, obj) for p in self.perms)

                cache[cache_key] = has_perm

            if has_perm:
                yield obj


class HasSourcePerm(HasPerm):
    """Defines permissions required to access the given field at object level.

    This will check the permissions for the source object to access the given field.

    Unlike `HasRetvalPerm`, this uses the source value (the object where the field
    is defined) to resolve the field, which means that this cannot be used for source
    queries and types.

    Examples
    --------
        To indicate that a field inside a `ProductType` can only be accessed if
        the user has "product.view_field" in it in the django system:

        >>> @gql.django.type(Product)
        ... class ProductType:
        ...     some_field: str = strawberry.field(
        ...         directives=[HasSourcePerm(".add_product")],
        ...     )

    """

    DEFAULT_TARGET: ClassVar[PermTarget] = PermTarget.SOURCE
    SCHEMA_DIRECTIVE_DESCRIPTION: ClassVar[Optional[str]] = _desc(
        "Will check if the user has any/all permissions for the parent "
        "of this field to resolve this.",
    )


class HasRetvalPerm(HasPerm):
    """Defines permissions required to access the given object/field at object level.

    Given a `app` name, the user can access the decorated object/field
    if he has any of the permissions defined in this directive.

    Note that this depends on resolving the object to check the permissions
    specifically for that object, unlike `HasPerm` which checks it before resolving.

    Examples
    --------
        To indicate that a field that returns a `ProductType` can only be accessed
        by someone who has "product.view_product"
        has "product.view_product" perm in the django system:

        >>> @strawberry.type
        ... class SomeType:
        ...     product: ProductType = strawberry.field(
        ...         directives=[HasRetvalPerm(".add_product")],
        ...     )

    """

    DEFAULT_TARGET: ClassVar[PermTarget] = PermTarget.RETVAL
    SCHEMA_DIRECTIVE_DESCRIPTION: ClassVar[Optional[str]] = _desc(
        "Will check if the user has any/all permissions for the resolved "
        "value of this field before returning it.",
    )



from django.http.request import HttpRequest
from strawberry.types import Info


def get_request(info: Info) -> HttpRequest:
    """Return the request from Info.

    description:
    Return the request object for both WSGI and ASGI implementations.
    It tends to move based on the environment.
    """
    try:
        request = info.context.request
    except AttributeError:
        request = info.context.get("request")

    return request



"""Code for interacting with Django settings."""

from typing import Optional, cast

from django.conf import settings
from typing_extensions import TypedDict


class StrawberryDjangoSettings(TypedDict):
    """Dictionary defining the shape `settings.STRAWBERRY_DJANGO` should have.

    All settings are optional and have defaults as described in their docstrings and
    defined in `DEFAULT_DJANGO_SETTINGS`.
    """

    #: If True, field descriptions will be fetched from the
    #: corresponding model field's `help_text` attribute.
    FIELD_DESCRIPTION_FROM_HELP_TEXT: bool

    #: If True, type descriptions will be fetched from the
    #: corresponding model model's docstring.
    TYPE_DESCRIPTION_FROM_MODEL_DOCSTRING: bool

    #: If True, fields with `choices` will have automatically generate
    #: an enum of possibilities instead of being exposed as `String`
    GENERATE_ENUMS_FROM_CHOICES: bool

    #: Set a custom default name for CUD mutations input type.
    MUTATIONS_DEFAULT_ARGUMENT_NAME: str

    #: If True, mutations will default to handling django errors by default
    #: when no option is passed to the field itself.
    MUTATIONS_DEFAULT_HANDLE_ERRORS: bool

    #: If True, `auto` fields that refer to model ids will be mapped to
    #: `relay.GlobalID` instead of `strawberry.ID` for types and filters.
    MAP_AUTO_ID_AS_GLOBAL_ID: bool

    #: Set a primary key default field name for Django CRUD resolvers.
    DEFAULT_PK_FIELD_NAME: str

    #: If True, deprecated way of using filters will be working
    USE_DEPRECATED_FILTERS: bool

    #: The default limit for pagination when not provided. Can be set to `None`
    #: to set it to unlimited.
    PAGINATION_DEFAULT_LIMIT: Optional[int]


DEFAULT_DJANGO_SETTINGS = StrawberryDjangoSettings(
    FIELD_DESCRIPTION_FROM_HELP_TEXT=False,
    TYPE_DESCRIPTION_FROM_MODEL_DOCSTRING=False,
    GENERATE_ENUMS_FROM_CHOICES=False,
    MUTATIONS_DEFAULT_ARGUMENT_NAME="data",
    MUTATIONS_DEFAULT_HANDLE_ERRORS=False,
    MAP_AUTO_ID_AS_GLOBAL_ID=False,
    DEFAULT_PK_FIELD_NAME="pk",
    USE_DEPRECATED_FILTERS=False,
    PAGINATION_DEFAULT_LIMIT=100,
)


def strawberry_django_settings() -> StrawberryDjangoSettings:
    """Get strawberry django settings.

    Return the dictionary from `settings.STRAWBERRY_DJANGO`, with defaults
    for missing keys.

    Preferred to direct access for the type hints and defaults.
    """
    defaults = DEFAULT_DJANGO_SETTINGS
    return cast(
        "StrawberryDjangoSettings",
        {**defaults, **getattr(settings, "STRAWBERRY_DJANGO", {})},
    )



import dataclasses
from collections.abc import Callable, Mapping, Sequence
from typing import (
    Any,
    Optional,
    TypeVar,
    Union,
    overload,
)

from strawberry.annotation import StrawberryAnnotation
from strawberry.extensions.field_extension import FieldExtension
from strawberry.field_extensions import InputMutationExtension
from strawberry.permission import BasePermission
from strawberry.types.fields.resolver import StrawberryResolver
from strawberry.types.unset import UNSET, UnsetType
from typing_extensions import Literal

from .fields import (
    DjangoCreateMutation,
    DjangoDeleteMutation,
    DjangoMutationBase,
    DjangoUpdateMutation,
)
from .types import FullCleanOptions

_T = TypeVar("_T")


@overload
def mutation(
    *,
    resolver: Callable[[], _T],
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[False] = False,
    permission_classes: Optional[list[type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    graphql_type: Optional[Any] = None,
    extensions: Optional[list[FieldExtension]] = None,
    handle_django_errors: Optional[bool] = None,
) -> _T: ...


@overload
def mutation(
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[True] = True,
    permission_classes: Optional[list[type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    graphql_type: Optional[Any] = None,
    extensions: Optional[list[FieldExtension]] = None,
    handle_django_errors: Optional[bool] = None,
) -> Any: ...


@overload
def mutation(
    resolver: Union[StrawberryResolver, Callable, staticmethod, classmethod],
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[list[type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    graphql_type: Optional[Any] = None,
    extensions: Optional[list[FieldExtension]] = None,
    handle_django_errors: Optional[bool] = None,
) -> DjangoMutationBase: ...


def mutation(
    resolver=None,
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[list[type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    graphql_type: Optional[Any] = None,
    extensions: Optional[list[FieldExtension]] = None,
    handle_django_errors: Optional[bool] = None,
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Optional[bool] = None,
) -> Any:
    """Annotate a property or a method to create a mutation field."""
    f = DjangoMutationBase(
        python_name=None,
        django_name=field_name,
        graphql_name=name,
        type_annotation=StrawberryAnnotation.from_annotation(graphql_type),
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives,
        extensions=extensions or (),
        handle_django_errors=handle_django_errors,
    )

    if resolver is not None:
        return f(resolver)

    return f


@overload
def input_mutation(
    *,
    resolver: Callable[[], _T],
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[False] = False,
    permission_classes: Optional[list[type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    graphql_type: Optional[Any] = None,
    extensions: Optional[list[FieldExtension]] = None,
    handle_django_errors: Optional[bool] = None,
) -> _T: ...


@overload
def input_mutation(
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[True] = True,
    permission_classes: Optional[list[type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    graphql_type: Optional[Any] = None,
    extensions: Optional[list[FieldExtension]] = None,
    handle_django_errors: Optional[bool] = None,
) -> Any: ...


@overload
def input_mutation(
    resolver: Union[StrawberryResolver, Callable, staticmethod, classmethod],
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[list[type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    graphql_type: Optional[Any] = None,
    extensions: Optional[list[FieldExtension]] = None,
    handle_django_errors: Optional[bool] = None,
) -> DjangoMutationBase: ...


def input_mutation(
    resolver=None,
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[list[type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    graphql_type: Optional[Any] = None,
    extensions: Optional[list[FieldExtension]] = None,
    handle_django_errors: Optional[bool] = None,
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Optional[bool] = None,
) -> Any:
    """Annotate a property or a method to create an input mutation field."""
    extensions = [*(extensions or []), InputMutationExtension()]
    f = DjangoMutationBase(
        python_name=None,
        django_name=field_name,
        graphql_name=name,
        type_annotation=StrawberryAnnotation.from_annotation(graphql_type),
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives,
        extensions=extensions,
        handle_django_errors=handle_django_errors,
    )

    if resolver is not None:
        return f(resolver)

    return f


def create(
    input_type: Optional[type] = None,
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[True] = True,
    permission_classes: Optional[list[type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    graphql_type: Optional[Any] = None,
    extensions: Optional[list[FieldExtension]] = None,
    argument_name: Optional[str] = None,
    handle_django_errors: Optional[bool] = None,
    full_clean: Union[bool, FullCleanOptions] = True,
) -> Any:
    """Create mutation for django input fields.

    Automatically create data for django input fields.

    Examples
    --------
        >>> @strawberry.django.input
        ... class ProductInput:
        ...     name: strawberry.auto
        ...     price: strawberry.auto
        ...
        >>> @strawberry.mutation
        >>> class Mutation:
        ...     create_product: ProductType = strawberry.django.create_mutation(
        ...         ProductInput
        ...     )

    """
    return DjangoCreateMutation(
        input_type,
        python_name=None,
        django_name=field_name,
        graphql_name=name,
        type_annotation=StrawberryAnnotation.from_annotation(graphql_type),
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives,
        extensions=extensions or (),
        argument_name=argument_name,
        handle_django_errors=handle_django_errors,
        full_clean=full_clean,
    )


def update(
    input_type: Optional[type] = None,
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    filters: Union[type, UnsetType, None] = UNSET,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[True] = True,
    permission_classes: Optional[list[type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    graphql_type: Optional[Any] = None,
    extensions: Optional[list[FieldExtension]] = None,
    argument_name: Optional[str] = None,
    handle_django_errors: Optional[bool] = None,
    key_attr: Optional[str] = None,
    full_clean: Union[bool, FullCleanOptions] = True,
) -> Any:
    """Update mutation for django input fields.

    Examples
    --------
        >>> @strawberry.django.input
        ... class ProductInput(IdInput):
        ...     name: strawberry.auto
        ...     price: strawberry.auto
        ...
        >>> @strawberry.mutation
        >>> class Mutation:
        ...     create_product: ProductType = strawberry.django.update_mutation(
        ...         ProductInput
        ...     )

    """
    return DjangoUpdateMutation(
        input_type,
        python_name=None,
        django_name=field_name,
        graphql_name=name,
        type_annotation=StrawberryAnnotation.from_annotation(graphql_type),
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives,
        filters=filters,
        extensions=extensions or (),
        argument_name=argument_name,
        handle_django_errors=handle_django_errors,
        key_attr=key_attr,
        full_clean=full_clean,
    )


def delete(
    input_type: Optional[type] = None,
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    filters: Union[type, UnsetType, None] = UNSET,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[True] = True,
    permission_classes: Optional[list[type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    extensions: Optional[list[FieldExtension]] = None,
    graphql_type: Optional[Any] = None,
    argument_name: Optional[str] = None,
    handle_django_errors: Optional[bool] = None,
    key_attr: Optional[str] = None,
    full_clean: Union[bool, FullCleanOptions] = True,
) -> Any:
    return DjangoDeleteMutation(
        input_type=input_type,
        python_name=None,
        django_name=field_name,
        graphql_name=name,
        type_annotation=StrawberryAnnotation.from_annotation(graphql_type),
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives,
        filters=filters,
        extensions=extensions or (),
        argument_name=argument_name,
        handle_django_errors=handle_django_errors,
        key_attr=key_attr,
        full_clean=full_clean,
    )



from typing import Literal, overload

from asgiref.sync import sync_to_async
from strawberry.types import Info

from strawberry_django.utils.requests import get_request
from strawberry_django.utils.typing import UserType


@overload
def get_current_user(info: Info, *, strict: Literal[True]) -> UserType: ...


@overload
def get_current_user(info: Info, *, strict: bool = False) -> UserType: ...


def get_current_user(info: Info, *, strict: bool = False) -> UserType:
    """Get and return the current user based on various scenarios."""
    request = get_request(info)

    try:
        user = request.user
    except AttributeError:
        try:
            # queries/mutations in ASGI move the user into consumer scope
            user = request.consumer.scope["user"]  # type: ignore
        except AttributeError:
            # websockets / subscriptions move scope inside of the request
            user = request.scope.get("user")  # type: ignore

    if user is None:
        raise ValueError("No user found in the current request")

    # Access an attribute inside the user object to force loading it in async contexts.
    _ = user.is_authenticated

    return user


@overload
async def aget_current_user(
    info: Info,
    *,
    strict: Literal[True],
) -> UserType: ...


@overload
async def aget_current_user(
    info: Info,
    *,
    strict: bool = False,
) -> UserType: ...


async def aget_current_user(info: Info, *, strict: bool = False) -> UserType:
    return await sync_to_async(get_current_user)(info, strict=strict)



import json
from dataclasses import dataclass
from json import JSONDecodeError
from typing import Any, ClassVar, Optional, cast

import strawberry
from asgiref.sync import sync_to_async
from django.core.exceptions import ValidationError
from django.db import DEFAULT_DB_ALIAS, models
from django.db.models import Expression, F, OrderBy, Q, QuerySet, Value, Window
from django.db.models.constants import LOOKUP_SEP
from django.db.models.expressions import Col
from django.db.models.functions import RowNumber
from django.db.models.sql.datastructures import BaseTable
from strawberry import Info, relay
from strawberry.relay import NodeType, PageInfo, from_base64
from strawberry.relay.types import NodeIterableType
from strawberry.relay.utils import should_resolve_list_connection_edges
from strawberry.types import get_object_definition
from strawberry.types.base import StrawberryContainer
from strawberry.utils.await_maybe import AwaitableOrValue
from strawberry.utils.inspect import in_async_context
from typing_extensions import Self

from strawberry_django.pagination import apply_window_pagination, get_total_count
from strawberry_django.queryset import get_queryset_config
from strawberry_django.resolvers import django_resolver


def _get_order_by(qs: QuerySet) -> list[OrderBy]:
    return [
        expr
        for expr, _ in qs.query.get_compiler(
            using=qs._db or DEFAULT_DB_ALIAS  # type: ignore
        ).get_order_by()
    ]


@dataclass
class OrderingDescriptor:
    attname: str
    order_by: OrderBy
    # we have to assume everything is nullable by default
    maybe_null: bool = True

    def get_comparator(self, value: Any, before: bool) -> Optional[Q]:
        if value is None:
            # 1. When nulls are first:
            #    1.1 there is nothing before "null"
            #    1.2 after "null" comes everything non-null
            # 2. When nulls are last:
            #    2.1 there is nothing after "null"
            #    2.2 before "null" comes everything non-null
            # => 1.1 and 2.1 require no checks
            # => 1.2 and 2.2 require an "is not null" check
            if bool(self.order_by.nulls_first) ^ before:
                return Q((f"{self.attname}{LOOKUP_SEP}isnull", False))
            return None
        lookup = "lt" if before ^ self.order_by.descending else "gt"
        cmp = Q((f"{self.attname}{LOOKUP_SEP}{lookup}", value))

        if self.maybe_null and bool(self.order_by.nulls_first) == before:
            # if nulls are first, "before any value" can also mean "is null"
            # if nulls are last, "after any value" can also mean "is null"
            cmp |= Q((f"{self.attname}{LOOKUP_SEP}isnull", True))
        return cmp

    def get_eq(self, value) -> Q:
        if value is None:
            return Q((f"{self.attname}{LOOKUP_SEP}isnull", True))
        return Q((f"{self.attname}{LOOKUP_SEP}exact", value))


def annotate_ordering_fields(
    qs: QuerySet,
) -> tuple[QuerySet, list[OrderingDescriptor], list[OrderBy]]:
    annotations = {}
    descriptors = []
    new_defer = None
    new_only = None
    order_bys = _get_order_by(qs)
    pk_in_order = False
    for index, order_by in enumerate(order_bys):
        if isinstance(order_by.expression, Col) and isinstance(
            # Col.alias is missing from django-types
            qs.query.alias_map[order_by.expression.alias],  # type: ignore
            BaseTable,
        ):
            field_name = order_by.expression.field.name
            # if it's a field in the base table, just make sure it is not deferred (e.g. by the optimizer)
            existing, defer = qs.query.deferred_loading
            if defer and field_name in existing:
                # Query is in "defer fields" mode and our field is being deferred
                if new_defer is None:
                    new_defer = set(existing)
                new_defer.discard(field_name)
            elif not defer and field_name not in existing:
                # Query is in "only these fields" mode and our field is not in the set of fields
                if new_only is None:
                    new_only = set(existing)
                new_only.add(field_name)
            descriptors.append(
                OrderingDescriptor(
                    order_by.expression.field.attname,
                    order_by,
                    maybe_null=order_by.expression.field.null,
                )
            )
            if order_by.expression.field.primary_key:
                pk_in_order = True
        else:
            dynamic_field = f"_strawberry_order_field_{index}"
            annotations[dynamic_field] = order_by.expression
            descriptors.append(OrderingDescriptor(dynamic_field, order_by))

    if new_defer is not None:
        # defer is additive, so clear it first
        qs = qs.defer(None).defer(*new_defer)
    elif new_only is not None:
        # only overwrites
        qs = qs.only(*new_only)

    if not pk_in_order:
        # Ensure we always have a clearly defined order by ordering by pk if it isn't in the order already
        # We cannot use QuerySet.order_by, because it validates the order expressions again,
        # but we're operating on the OrderBy expressions which have already been resolved by the compiler
        # In case the user has previously ordered by an aggregate like so:
        # qs.annotate(_c=Count("foo")).order_by("_c")  # noqa: ERA001
        # then the OrderBy we get here would trigger a ValidationError by QuerySet.order_by.
        # But we only want to append to the existing order (and the existing order must be valid already)
        # So this is safe.
        pk_order = F("pk").resolve_expression(qs.query).asc()
        order_bys.append(pk_order)
        descriptors.append(OrderingDescriptor("pk", pk_order, maybe_null=False))
        qs = qs._chain()  # type: ignore
        qs.query.order_by += (pk_order,)
    return qs.annotate(**annotations), descriptors, order_bys


def build_tuple_compare(
    descriptors: list[OrderingDescriptor],
    cursor_values: list[Optional[str]],
    before: bool,
) -> Q:
    current = None
    for descriptor, field_value in zip(reversed(descriptors), reversed(cursor_values)):
        if field_value is None:
            value_expr = None
        else:
            output_field = descriptor.order_by.expression.output_field
            value_expr = Value(field_value, output_field=output_field)
        cmp = descriptor.get_comparator(value_expr, before)
        if current is None:
            current = cmp
        else:
            eq = descriptor.get_eq(value_expr)
            current = cmp | (eq & current) if cmp is not None else eq & current
    return current if current is not None else Q()


class AttrHelper:
    pass


def _extract_expression_value(
    model: models.Model, expr: Expression, attname: str
) -> Optional[str]:
    output_field = expr.output_field
    # Unfortunately Field.value_to_string operates on the object, not a direct value
    # So we have to potentially construct a fake object
    # If the output field's attname doesn't match, we have to construct a fake object
    # Additionally, the output field may not have an attname at all
    # if expressions are used
    field_attname = getattr(output_field, "attname", None)
    if not field_attname:
        # If the field doesn't have an attname, it's a dynamically constructed field,
        # for the purposes of output_field in an expression. Just set its attname, it doesn't hurt anything
        output_field.attname = field_attname = attname
    obj: Any
    if field_attname != attname:
        obj = AttrHelper()
        setattr(obj, output_field.attname, getattr(model, attname))
    else:
        obj = model
    value = output_field.value_from_object(obj)
    if value is None:
        return None
    # value_to_string is missing from django-types
    return output_field.value_to_string(obj)  # type: ignore


def apply_cursor_pagination(
    qs: QuerySet,
    *,
    related_field_id: Optional[str] = None,
    info: Info,
    before: Optional[str],
    after: Optional[str],
    first: Optional[int],
    last: Optional[int],
    max_results: Optional[int],
) -> tuple[QuerySet, list[OrderingDescriptor]]:
    max_results = (
        max_results if max_results is not None else info.schema.config.relay_max_results
    )

    qs, ordering_descriptors, original_order_by = annotate_ordering_fields(qs)
    if after:
        after_cursor = OrderedCollectionCursor.from_cursor(after, ordering_descriptors)
        qs = qs.filter(
            build_tuple_compare(ordering_descriptors, after_cursor.field_values, False)
        )
    if before:
        before_cursor = OrderedCollectionCursor.from_cursor(
            before, ordering_descriptors
        )
        qs = qs.filter(
            build_tuple_compare(ordering_descriptors, before_cursor.field_values, True)
        )

    slice_: Optional[slice] = None
    if first is not None and last is not None:
        if last > max_results:
            raise ValueError(f"Argument 'last' cannot be higher than {max_results}.")
        # if first and last are given, we have to
        # - reverse the order in the DB so we can use slicing to apply [:last],
        #   otherwise we would have to know the total count to apply slicing from the end
        # - We still need to apply forward-direction [:first] slicing, and according to the Relay spec,
        #   it shall happen before [:last] slicing. To do this, we use a window function with a RowNumber ordered
        #   in the original direction, which is opposite the actual query order.
        #   This query is likely not very efficient, but using last _and_ first together is discouraged by the
        #   spec anyway
        qs = (
            qs.reverse()
            .annotate(
                _strawberry_row_number_fwd=Window(
                    RowNumber(),
                    order_by=original_order_by,
                ),
            )
            .filter(
                _strawberry_row_number_fwd__lte=first + 1,
            )
        )
        # we're overfetching by two, in both directions
        slice_ = slice(last + 2)
    elif first is not None:
        if first < 0:
            raise ValueError("Argument 'first' must be a non-negative integer.")
        if first > max_results:
            raise ValueError(f"Argument 'first' cannot be higher than {max_results}.")
        slice_ = slice(first + 1)
    elif last is not None:
        # when using last, optimize by reversing the QuerySet ordering in the DB,
        # then slicing from the end (which is now the start in QuerySet ordering)
        # and then iterating the results in reverse to restore the original order
        if last < 0:
            raise ValueError("Argument 'last' must be a non-negative integer.")
        if last > max_results:
            raise ValueError(f"Argument 'last' cannot be higher than {max_results}.")
        slice_ = slice(last + 1)
        qs = qs.reverse()
    if related_field_id is not None:
        # we always apply window pagination for nested connections,
        # because we want its total count annotation
        offset = slice_.start or 0 if slice_ is not None else 0
        qs = apply_window_pagination(
            qs,
            related_field_id=related_field_id,
            offset=offset,
            limit=slice_.stop - offset if slice_ is not None else None,
        )
    elif slice_ is not None:
        qs = qs[slice_]

    get_queryset_config(qs).ordering_descriptors = ordering_descriptors

    return qs, ordering_descriptors


@dataclass
class OrderedCollectionCursor:
    field_values: list[Any]

    @classmethod
    def from_model(
        cls, model: models.Model, descriptors: list[OrderingDescriptor]
    ) -> Self:
        values = [
            _extract_expression_value(
                model, descriptor.order_by.expression, descriptor.attname
            )
            for descriptor in descriptors
        ]
        return cls(field_values=values)

    @classmethod
    def from_cursor(cls, cursor: str, descriptors: list[OrderingDescriptor]) -> Self:
        type_, values_json = from_base64(cursor)
        if type_ != DjangoCursorEdge.CURSOR_PREFIX:
            raise ValueError("Invalid cursor")
        try:
            string_values = json.loads(values_json)
        except JSONDecodeError as e:
            raise ValueError("Invalid cursor") from e
        if (
            not isinstance(string_values, list)
            or len(string_values) != len(descriptors)
            or any(not (v is None or isinstance(v, str)) for v in string_values)
        ):
            raise ValueError("Invalid cursor")

        try:
            decoded_values = [
                d.order_by.expression.output_field.to_python(v)
                for d, v in zip(descriptors, string_values)
            ]
        except ValidationError as e:
            raise ValueError("Invalid cursor") from e

        return cls(decoded_values)

    def __str__(self):
        return json.dumps(self.field_values, separators=(",", ":"))


@strawberry.type(name="CursorEdge", description="An edge in a connection.")
class DjangoCursorEdge(relay.Edge[relay.NodeType]):
    CURSOR_PREFIX: ClassVar[str] = "orderedcursor"


@strawberry.type(
    name="CursorConnection", description="A connection to a list of items."
)
class DjangoCursorConnection(relay.Connection[relay.NodeType]):
    total_count_qs: strawberry.Private[Optional[QuerySet]] = None
    edges: list[DjangoCursorEdge[NodeType]] = strawberry.field(  # type: ignore
        description="Contains the nodes in this connection"
    )

    @strawberry.field(description="Total quantity of existing nodes.")
    @django_resolver
    def total_count(self) -> int:
        assert self.total_count_qs is not None

        return get_total_count(self.total_count_qs)

    @classmethod
    def resolve_connection(
        cls,
        nodes: NodeIterableType[NodeType],
        *,
        info: Info,
        before: Optional[str] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        last: Optional[int] = None,
        max_results: Optional[int] = None,
        **kwargs: Any,
    ) -> AwaitableOrValue[Self]:
        from strawberry_django.optimizer import is_optimized_by_prefetching

        if not isinstance(nodes, QuerySet):
            raise TypeError("DjangoCursorConnection requires a QuerySet")
        total_count_qs: QuerySet = nodes
        qs: QuerySet
        if not is_optimized_by_prefetching(nodes):
            qs, ordering_descriptors = apply_cursor_pagination(
                nodes,
                info=info,
                before=before,
                after=after,
                first=first,
                last=last,
                max_results=max_results,
            )
        else:
            qs = nodes
            ordering_descriptors = get_queryset_config(qs).ordering_descriptors
            assert ordering_descriptors is not None

        type_def = get_object_definition(cls)
        assert type_def
        field_def = type_def.get_field("edges")
        assert field_def

        field = field_def.resolve_type(type_definition=type_def)
        while isinstance(field, StrawberryContainer):
            field = field.of_type

        edge_class = cast("DjangoCursorEdge[NodeType]", field)

        if not should_resolve_list_connection_edges(info):
            return cls(
                edges=[],
                total_count_qs=total_count_qs,
                page_info=PageInfo(
                    start_cursor=None,
                    end_cursor=None,
                    has_previous_page=False,
                    has_next_page=False,
                ),
            )

        def finish_resolving():
            nonlocal qs
            has_previous_page = has_next_page = False

            results = list(qs)

            if first is not None:
                if last is None:
                    has_next_page = len(results) > first
                    results = results[:first]
                # we're paginating forwards _and_ backwards
                # remove the (potentially) overfetched row in the forwards direction first
                elif (
                    results
                    and getattr(results[0], "_strawberry_row_number_fwd", 0) > first
                ):
                    has_next_page = True
                    results = results[1:]

            if last is not None:
                has_previous_page = len(results) > last
                results = results[:last]

            it = reversed(results) if last is not None else results

            edges = [
                edge_class.resolve_edge(
                    cls.resolve_node(v, info=info, **kwargs),
                    cursor=OrderedCollectionCursor.from_model(v, ordering_descriptors),
                )
                for v in it
            ]

            return cls(
                edges=edges,
                total_count_qs=total_count_qs,
                page_info=PageInfo(
                    start_cursor=edges[0].cursor if edges else None,
                    end_cursor=edges[-1].cursor if edges else None,
                    has_previous_page=has_previous_page,
                    has_next_page=has_next_page,
                ),
            )

        if in_async_context() and qs._result_cache is None:  # type: ignore
            return sync_to_async(finish_resolving)()
        return finish_resolving()
